# Make a sequence of colors that are each as different as can be from the previous, using different metrics and biases, as desired.
# Eric Haines, erich@acm.org
# Version 1.1, November 1, 2022
use strict;
use warnings;
use POSIX qw/ceil/;
$| = 1;	# flush text to screen immediately

my $findcolors = 100;	# how many colors to find.
my $divisions = 255;	# how much to subdivide the color cube, minus 1; i.e. 4 means 0.0, 0.25, 0.5, 0.75, 1.0 divisions
						# 255 takes overnight, 100 takes maybe an hour or so and is likely fine
my $non_gray = 0;		# To avoid grays, set to 0.15 or somesuch
my $min_sum = 0;		# To avoid dark colors, set to 0.2 or somesuch
my $favorlast = 0;		# If true (set to 1), do more to avoid being near the previous three colors generated
my $colorcube = 1;		# 1 - use linear RGB for distances; 0 - use CIELAB distances

my $outputsize = 32;	# size of color tiles output to PPM file
my $xres = 10;			# number of colors per row
my $yres = ceil($findcolors / $xres);	# no touch!

# output file names
my $fileroot = 'cs';
my $imgfilename = $fileroot . '_div' . $divisions . '_cc' . $colorcube . '_nograys' . $non_gray . '_minsum' . $min_sum . '_favor' . $favorlast . '.ppm';
my $txtfilename = $fileroot . '_div' . $divisions . '_cc' . $colorcube . '_nograys' . $non_gray . '_minsum' . $min_sum . '_favor' . $favorlast . '.txt';

my $gamma = 1/2.2;

#------------------------


#seed color, to start
my @reds;
my @greens;
my @blues;
#green
#$reds[0] = 0;
#$greens[0] = 1;
#$blues[0] = 0;
#gray
#$reds[0] = $greens[0] = $blues[0] = 0.5;
#white
$reds[0] = $greens[0] = $blues[0] = 1;

my @redslab;
my @greenslab;
my @blueslab;

($redslab[0],$greenslab[0],$blueslab[0]) = &linearRGBtoCIELAB($reds[0], $greens[0], $blues[0]);

my @colorset;
my @rgbset;

open(TFH, '>', $txtfilename) or die $!;

print "$txtfilename\n";
print TFH "$txtfilename\n";

my $i;
for ($i=0;$i<$findcolors;$i++) {
	my $farthest = -1;

	my ($r,$g,$b);
	if ( $i > 0 ) {
		#print "color $i\n";
		for ($r=0;$r<=$divisions;$r++) {
			for ($g=0;$g<=$divisions;$g++) {
				for ($b=0;$b<=$divisions;$b++) {
					my $distance = &colorDistance($r/$divisions,$g/$divisions,$b/$divisions,$i);
					# We want to take the largest closest distance.
					if ($distance > $farthest) {
						$farthest = $distance;
						$reds[$i] = $r/$divisions;
						$greens[$i] = $g/$divisions;
						$blues[$i] = $b/$divisions;
	#print ">> farthest is now $farthest for $reds[$i],$greens[$i],$blues[$i]\n";
					}
				}
			}
		}
	}

	# Save each out, converting to display colors. Just linearizing with gamma. Could use sRGBtoLinearRGB, but that's fiddling.
	$colorset[$i][0] = $reds[$i]**$gamma;
	$colorset[$i][1] = $greens[$i]**$gamma;
	$colorset[$i][2] = $blues[$i]**$gamma;
	if ( $colorcube == 0 ) {
		($redslab[$i],$greenslab[$i],$blueslab[$i]) = &linearRGBtoCIELAB($reds[$i], $greens[$i], $blues[$i]);
	}
	$rgbset[$i][0] = int(255*$colorset[$i][0]+0.5);
	$rgbset[$i][1] = int(255*$colorset[$i][1]+0.5);
	$rgbset[$i][2] = int(255*$colorset[$i][2]+0.5);
	printf "Color %2d is (%0.4f, %0.4f, %0.4f), int: (%3d, %3d, %3d)\n",
		$i, $colorset[$i][0], $colorset[$i][1], $colorset[$i][2],
		$rgbset[$i][0], $rgbset[$i][1], $rgbset[$i][2];
	printf TFH "Color %2d is (%0.4f, %0.4f, %0.4f), int: (%3d, %3d, %3d)\n",
		$i, $colorset[$i][0], $colorset[$i][1], $colorset[$i][2],
		$rgbset[$i][0], $rgbset[$i][1], $rgbset[$i][2];
}

close TFH;

open(IFH, '>', $imgfilename) or die $!;

print IFH "P3\n";

printf IFH "%d %d\n255\n", $xres*$outputsize, $yres*$outputsize;

# To instead store to an array which is then output:
#my @img = ("0 0 0 ") x ($xres*$outputsize*$yres*$outputsize);

# TODO: a little slower overall, but could make the PPM after each color is found,
# writing over it each time. For the impatient.
my $colornum = 0;
for (my $y=0;$y<$yres;$y++) {
	for (my $yl = 0; $yl < $outputsize; $yl++){
		$colornum = $y * $xres;
		for (my $x=0;$x<$xres;$x++) {
			# is color to be output?
			if ( $colornum < $findcolors ) {
				my $xloc = $x * $outputsize;
				my $yloc = int($colornum / $xres) * $outputsize;
				my $loc = $x * $outputsize + $yloc * $outputsize * $xres * $outputsize;

				my $color = sprintf "%d %d %d ", $rgbset[$colornum][0], $rgbset[$colornum][1], $rgbset[$colornum][2];
				for (my $xl = 0; $xl < $outputsize; $xl++){
					#$img[$loc+$yl*$xres*$outputsize+$xl] = $color;
					print IFH $color;
				}
				$colornum++;
			} else {
				# no color output
				for (my $xl = 0; $xl < $outputsize; $xl++){
					print IFH "0 0 0 ";
				}
			}
		}
		printf IFH "\n";
	}
}

close IFH;

#===========================
	
# returns distance of -1 if color is rejected for some reason
sub colorDistance {
	my ($r,$g,$b,$cidx) = @_;
	
	# if color is a gray, don't use it
	if (abs($r-$g) < $non_gray && abs($r-$b) < $non_gray && abs($g-$b) < $non_gray) {
		# nearly gray
		return -1;
	}

	# if color is too dim, don't use it
	if ($r+$g+$b < $min_sum) {
		return -1;
	}
	my ($inr,$ing,$inb);
	if ( $colorcube == 0 ) {
		($inr,$ing,$inb) = &linearRGBtoCIELAB($r,$g,$b);
	}

	my $max_distance = 99999;
	# go through previous colors and find closest one
	for (my $c=0;$c<$cidx;$c++){
		my $dist;
		if ( $colorcube == 1 ) {
			$dist = ($r-$reds[$c])*($r-$reds[$c]) + ($g-$greens[$c])*($g-$greens[$c]) + ($b-$blues[$c])*($b-$blues[$c]);
		} else {
			$dist = ($inr-$redslab[$c])*($inr-$redslab[$c]) + ($ing-$greenslab[$c])*($ing-$greenslab[$c]) + ($inb-$blueslab[$c])*($inb-$blueslab[$c]);
		}

		if ( $favorlast && $cidx - $c <= 3 ) {
			# For the last 3 colors, take the new color's distance from each and DECREASE it
			# by some small amount: 0.7, 0.8, 0.9 of the distance for these three, descending.
			# We're looking for the max distance away from all, so decreasing the distance for
			# these three will make neighbors to them be considered "too close."
			$dist /= (1 + 0.1*(4-$cidx+$c));
		}
#print "Dist is $dist for $r,$g,$b vs. $c: $reds[$c], $greens[$c], $blues[$c]\n";
		if ($dist < $max_distance) {
			# The max_distance for the test point is the minimum of how far it is from all previous colors
			$max_distance = $dist;
			if ( $dist == 0 ) {
				# means we've already used this color, so it definitely won't be The One - early out
				return -1;
			}
#print "max dist is now $max_distance\n";
		}
	}
	return $max_distance;
}

# not used, but just in case...
sub sRGBtoLinearRGB {
	# Code snippets from http://www.easyrgb.com/en/math.php
	#sR, sG and sB (Standard RGB) input range = 0 ÷ 255
	#X, Y and Z output refer to a D65/2° standard illuminant.
	my ($sR,$sG,$sB) = @_;

	my $var_R = ( $sR / 255 );
	my $var_G = ( $sG / 255 );
	my $var_B = ( $sB / 255 );

	# convert from sRGB to linear values
	if ( $var_R > 0.04045 ) { $var_R = ( ( $var_R + 0.055 ) / 1.055 ) ** 2.4; }
	else                   { $var_R = $var_R / 12.92; }
	if ( $var_G > 0.04045 ) { $var_G = ( ( $var_G + 0.055 ) / 1.055 ) ** 2.4; }
	else                   { $var_G = $var_G / 12.92; }
	if ( $var_B > 0.04045 ) { $var_B = ( ( $var_B + 0.055 ) / 1.055 ) ** 2.4; }
	else                   { $var_B = $var_B / 12.92; }
	
	return ($var_R, $var_G, $var_B);
}

sub linearRGBtoCIELAB {
	# https://github.com/NVlabs/flip/blob/main/cpp/CPP/color.h
	# linear RGB input
	#X, Y and Z output refer to a D65/2° standard illuminant.

	my ($var_R, $var_G, $var_B) = @_;

	# Source: https://www.image-engineering.de/library/technotes/958-how-to-convert-between-srgb-and-ciexyz
	# Assumes D65 standard illuminant.
	my $a11 = 10135552.0 / 24577794.0;
	my $a12 = 8788810.0 / 24577794.0;
	my $a13 = 4435075.0 / 24577794.0;
	my $a21 = 2613072.0 / 12288897.0;
	my $a22 = 8788810.0 / 12288897.0;
	my $a23 = 887015.0 / 12288897.0;
	my $a31 = 1425312.0 / 73733382.0;
	my $a32 = 8788810.0 / 73733382.0;
	my $a33 = 70074185.0 / 73733382.0;

	my $X = $a11 * $var_R + $a12 * $var_G + $a13 * $var_B;
	my $Y = $a21 * $var_R + $a22 * $var_G + $a23 * $var_B;
	my $Z = $a31 * $var_R + $a32 * $var_G + $a33 * $var_B;

	my $delta = 6.0 / 29.0;
	my $deltaSquare = $delta * $delta;
	my $deltaCube = $delta * $deltaSquare;
	my $factor = 1.0 / (3.0 * $deltaSquare);
	my $term = 4.0 / 29.0;

	# The default illuminant is D65.
	my @Reference = (1.052156925, 1.000000000, 0.918357670);
	my $var_X = $X / $Reference[0];
	my $var_Y = $Y / $Reference[1];
	my $var_Z = $Z / $Reference[2];

	$var_X = ($var_X > $deltaCube ? $var_X**(1.0 / 3.0) : $factor * $var_X + $term);
	$var_Y = ($var_Y > $deltaCube ? $var_Y**(1.0 / 3.0) : $factor * $var_Y + $term);
	$var_Z = ($var_Z > $deltaCube ? $var_Z**(1.0 / 3.0) : $factor * $var_Z + $term);
	my $L = 116.0 * $var_Y - 16.0;
	my $a = 500.0 * ($var_X - $var_Y);
	my $b = 200.0 * ($var_Y - $var_Z);
	
	return ($L, $a, $b);
}
